; ISMEMO.ASM
; ************ CORRECTED PER TECHNOTES CORRIGENDA IN MAY 1988 ISSUE *****
; Program..:  Ismemo.ASM  
; Author...:  Kenneth N. Getz  
; Date.....:  January 1, 1988  
; Notes....:  Ismemo.ASM tests for the existence of a memo field for a   
;             specific record in a specific file.  The routine is   
;             called with the following parameter string stored in a  
;             memory variable:  
;  
;                 STR(offset of the memo field, 4) + ;
;                 STR(record number, 10) +;  
;                 "database filename"  
;  
;             If no error occurs, the first character in the memory  
;             variable parameter will be a space (CHR(32)) and the   
;             second character will be either a "Y" or an "N", depending  
;             on whether a memo field was found.  
;  
;             If an error occurs, one of the following codes will be  
;             returned in the first byte of the memory variable.  The  
;             first 5 (2 - 6) are assigned by DOS, and the last four  
;             are assigned by Ismemo.  
;  
;             Errors:  
;                 CHR(2) : File not found.  
;                 CHR(3) : Path not found or file doesn't exist.  
;                 CHR(4) : No file handle available.  
;                 CHR(5) : Access denied.  
;                 CHR(6) : Handle invalid or not open.  
;                 CHR(7) : The file was too small.  
;                 CHR(8) : The record number was out of range.  
;                 CHR(9) : The record offset passed in was non-numeric.  
;                 CHR(10): The record number passed in was non-numeric.  
;                 Any other character : Unspecified file error.  
;   
;             You will find that the code for this program is very similar  
;             to the code found in Mem2char.ASM from the August 1987  
;             issue of TechNotes.  If you've already entered that program,   
;             you may want to make a copy of it and start editing from there.  
;  
;****************************************************************************  
DOS     EQU     21h                    ; Call to DOS.  
CREATE  EQU     3Ch                    ; Create a file.
OPEN    EQU     3Dh                    ; Open a file.  
CLOSE   EQU     3Eh                    ; Close a file.  
READ    EQU     3Fh                    ; Read from a file.  
WRITE   EQU     40h                    ; Write to a file.  
MOVE    EQU     42h                    ; Move file pointer.  
X_SIZE  EQU     7                      ; Indicates the file was too small.  
X_RNUM  EQU     8                      ; Indicates record number out of range.  
X_FLD   EQU     9                      ; Record offset non-numeric.  
X_REC   EQU     10                     ; Record number non-numeric.  
;-----------------------------------------------------------------------------  
SAVE    MACRO   R1,R2,R3,R4,R5,R6,R7,R8  
;  
; This macro takes a parameter list and pushes each parameter register, in the  
; order given, onto the stack.    
;  
;       USAGE:  SAVE AX,BX,CX...  
;
          IRP     X,<R1,R2,R3,R4,R5,R6,R7,R8>
                  IFNB    <X>
                          PUSH X
                  ENDIF
          ENDM
          ENDM      
;-----------------------------------------------------------------------------  
RESTORE MACRO   R1,R2,R3,R4,R5,R6,R7,R8  
;  
; This macro takes a parameter list and pops each parameter register, in the  
; opposite of the order given, from the stack.  
;  
;       USAGE:  RESTORE AX,BX,CX...  
;
          IRP     X,<R8,R7,R6,R5,R4,R3,R2,R1>
                  IFNB    <X>
                          POP X
                  ENDIF
          ENDM
          ENDM      
;-----------------------------------------------------------------------------
GETFILE MACRO   MODE  
LOCAL   L1,L2,L3  
;  
; This macro opens a file, putting the file handle into the variable HANDLE.  
; If an error occurs, the error code is sent back to dBASE in the first  
; character of the passed parameter.  The ASCIIZ string for the file  
; name is passed in by dBASE in the DS:DX pair.   
;  
;       USAGE:  GETFILE CREATE  
;       USAGE:  GETFILE READ  
;
          XOR     AX,AX                  ; Clear out AX.
          MOV     DX,OFFSET CS:FILENM    ; INT 21h needs address in DS:DX
          MOV     AH,MODE                ; Get the correct mode in AH.
          MOV     AL,2
          CLC
          INT     DOS                    ; Call DOS.
          JC      L2                     
                                         ; Check for carry flag set (ERROR!)  
L1:       MOV     CS:HANDLE,AX           ; If OK, move handle to storage.
          JMP     L3  
L2:       MOV     CS:ERROR,AX            ; Otherwise, store error code.
L3:       NOP
          ENDM  
;----------------------------------------------------------------------------  
DOSFILE MACRO   @AH,@AL,@BX,@CX,@DX  
;  
; This macro simplifies calling DOS for all interrupt functions.  
;  
;       USAGE:  DOSFILE READ,0,HANDLE,1,<OFFSET CHECK>  
;
          MOV     AH,@AH
          MOV     AL,@AL
          MOV     BX,@BX
          MOV     CX,@CX
          MOV     DX,@DX
          INT     DOS
          ENDM  
;****************************************************************************  
CODE    SEGMENT BYTE PUBLIC     'CODE'    
;----------------------------------------------------------------------------  
ISMEM     PROC    FAR
          ASSUME  CS:code
 
START:    JMP     ENTRY  ;============================================================================
          DB      13,'Ismemo.BIN',10,13
          DB      'Written by Kenneth N. Getz.',10,13
          DB      '(C) Copyright 1988, Ashton-Tate',10,13 
          DB      26
 
TEN       DW      10                     ; Constant used for multiplication.  
CHECK     DB      10 DUP(0)              ; Data area for memo field check.  
RECOFFS   DW      0                      ; Offset in record for memo field.  
RECPOS    DB      4  DUP(0)              ; ASCII offset for memo field.  
RECNO     DB      10 DUP(0)              ; ASCII record number.  
FILENM    DB      65 DUP(0)              ; Filename.  
HANDLE    DW      0                      ; File handle.  
ERROR     DW      0                      ; Error flag.  
POS1      DW      ?                      ; Low-word of the file offset.  
POS2      DW      ?                      ; High-word of the file offset.  
HEADER    STRUC                          ; Structure for header information.
          VERSION   DB      0
          DATE      DB      3 DUP(0)
          NUMRECS   DD      0
          HEADSZ    DW      0
          RECLEN    DW      0
          RESRVED   DB      20 DUP(0)
HEADER    ENDS
HEAD      HEADER  <>                     ; Initialization of structure.  
ISMEMO    DB      ' '                    ; Data storage - is there a memo field?  ;=============================================================================  
ENTRY:    SAVE    AX,BX,CX,DX,SI,DI      ; Save the environment.
          PUSH    DS                     ; Save dBASE's Data Segment.
          MOV     HANDLE,0               ; Initialize file handle to 0.
          MOV     ISMEMO,32              ; Initialize the memo flag.
          CALL    PARAMS                 ; Get the passed parameters.
          PUSH    CS                     ; Make DS = CS.
          POP     DS
          MOV     ERROR,32               ; Initialize error flag to be a space.
          GETFILE OPEN                   ; Open the necessary file.
          CMP     ERROR,32               ; See if an error occurred.
          JNE     GOUT                   ; If so, quit.
          CALL    GETHD                  ; Read in the file header.
          CMP     ERROR,32               ; See if an error occurred.
          JNE     GOUT                   ; If so, quit.
          CALL    MAKECHG                ; Do the actual work.
GOUT:     POP     DS                     ; Restore dBASE's DS register.
          CMP     CS:HANDLE,0            ; See if we ever opened a file.
          JZ      NO_CLS                 ; If not, don't close one.
          CALL    CLSFILE                ; Otherwise, close the file.  
NO_CLS:   MOV     AL,BYTE PTR CS:ERROR   ; Send back the error byte.
          MOV     [BX],AL                ; Tell dBASE about error.
          MOV     AL,BYTE PTR CS:ISMEMO  ; Was a memo field found?
          MOV     [BX+1],AL              ; Tell dBASE about it.
          RESTORE AX,BX,CX,DX,SI,DI      ; Restore the environment.
          RET                            ; And return to dBASE.
  ISMEM   ENDP  
;----------------------------------------------------------------------------  
PARAMS  PROC    NEAR  
;  
; This procedure fills the data area with arguments passed on the command  
; line from dBASE.  This routine is very dependent on the order of the  
; variables in the data area.
;  
          SAVE    AX,CX,DX,DS            ; Save the environment.
          MOV     CL,BYTE PTR [BX-1]     ; Get dBASE's param string length.
          XOR     CH,CH                  ; Clear the high byte.
          DEC     CX                     ; Don't want the last NULL byte.
          MOV     SI,BX                  ; Point SI to the string.
          MOV     DI,OFFSET RECPOS       ; Point DI to the buffer area.
          REP     MOVSB                  ; Copy it all in to the code area.
          XOR     AL,AL                  ; Zero out AL
          STOSB                          ; Make sure last byte is 0.
          RESTORE AX,CX,DX,DS            ; Restore the environment.
          RET               
PARAMS    ENDP
;----------------------------------------------------------------------------  
GETHD   PROC    NEAR  
;  
; This procedure reads the file header into the structure HEAD.  
; If an error occurs (32 bytes couldn't be read), ERROR will contain  
; X_SIZE.  
;
          SAVE    AX,BX,CX,DX            ; Read 32-byte file header.
          DOSFILE READ,0,HANDLE,32,<OFFSET HEAD>
          CMP     AX,CX                  ; See if we read all 32 bytes.
          JE      HD_OK                  ; If so, we're OK.
          MOV     ERROR,X_SIZE           ; Otherwise, we have an error.  
HD_OK:    RESTORE AX,BX,CX,DX            ; Restore all registers.
          RET  
GETHD     ENDP  
;----------------------------------------------------------------------------  
MAKECHG PROC    NEAR  
;  
; This procedure actually performs the work.  It reads the necessary byte  
;
          SAVE    AX,BX,CX,DX
          MOV     ISMEMO,32              ; Initialize it to be a space.
          MOV     SI,OFFSET RECNO        ; Point SI to the ASCII record number.
          CALL    CVRTDW                 ; Put record number in DX:AX.
          CMP     ERROR,32               ; Check for error.
          JNE     NO_CHG                 ; If there is one, quit.
          CMP     DX,WORD PTR HEAD.NUMRECS+2  ; DX better be less.
          JBE     HI_OK                  ; If DX OK, go on.
          MOV     ERROR,X_RNUM           ; Otherwise, set error.
          JMP     NO_CHG                 ; And quit.
  HI_OK:  CMP     AX,WORD PTR HEAD.NUMRECS    ; AX better be less.
          JBE     LO_OK                  ; IF AX OK, go on.
          MOV     ERROR,X_RNUM           ; Otherwise, set error.
          JMP     NO_CHG                 ; And quit.
  LO_OK:  SUB     AX,1
          SBB     DX,0
          MOV     DI,HEAD.RECLEN         ; Get the record length.
          CALL    MULT                   ; Multiply to get file offset.
          ADD     AX,HEAD.HEADSZ         ; Add on header size.
          ADC     DX,0                   ; Update high word if necessary.
          CALL    CVRT                   ; Puts memo offset into BX.
          CMP     ERROR,32               ; See if error occurred.
          JNE     NO_CHG                 ; If so, quit.
          ADD     AX,BX                  ; Add memo offset to file offset.
          ADC     DX,0                   ; Update high word if necessary.
          INC     AX                     ; Skip over first byte (' ').
          ADC     DX,0                   ; Update high word if necessary.
  SIZE_OK:
          MOV     POS2,DX
          MOV     POS1,AX                ; Store away the file position.
          DOSFILE MOVE,0,HANDLE,POS2,POS1          ; Go to that file offset.
          DOSFILE READ,0,HANDLE,10,<OFFSET CHECK>  ; Read the bytes.
          CALL    MEMTEST                ; Check for a memo field.
  NO_CHG:
          RESTORE AX,BX,CX,DX            ; Restore all registers.
          RET  
MAKECHG   ENDP  
;----------------------------------------------------------------------------
CLSFILE   PROC    NEAR  
;   
; This procedure closes the file referred to by HANDLE.  If an error  
; occurred while closing the file, the error code is passed back to  
; dBASE in the first character of the calling parameter.  
;
          SAVE    AX,BX
          MOV     BX,CS:HANDLE           ; Get the file handle.
          MOV     AH,CLOSE               ; Set up for Close.
          CLC                            ; Clear the Carry Flag.
          INT     DOS                    ; Call DOS.
          JNC     CLS_OK                 ; If Carry Flag still clear, we're OK.
          MOV     CS:ERROR,AX            ; Otherwise, send error back.
CLS_OK:
          RESTORE AX,BX
          RET
CLSFILE ENDP  
;----------------------------------------------------------------------------
CVRT    PROC    NEAR  
;  
; Ends up with string pointed to by SI converted to binary in BX.  
; This routine initializes BX to zero, and stops when it gets to any  
; non-numeric character.  This means that if you pass in a string like   
; "QRST", this procedure will return a 0 in BX.  If you pass in "123q",  
; it will place 123 in BX.  It does skip over leading spaces.  
;
          SAVE    AX,CX,DX
          SUB     AX,AX                  ; Clear out AX.
          SUB     BX,BX                  ; Clear out BX.
          MOV     SI,OFFSET RECPOS       ; Point SI to the parameter storage.
          MOV     CX,4                   ; Set loop counter.
NUMLP1:   CMP     BYTE PTR [SI],32       ; See if we got a space.
          JE      SPACE1                 ; If so, go on to next character.
          CMP     BYTE PTR [SI],'0'      ; See if character less than '0'.
          JB      ERR1                   ; If so, we have an error.
          CMP     BYTE PTR [SI],'9'      ; See if character greater than '9'.
          JA      ERR1                   ; If so, we have an error.
          MUL     TEN                    ; If no error, multiply AX by 10.
          MOV     BL,[SI]                ; Get the next input character.
          SUB     BX,'0'                 ; Subtract ASCII offset.
          ADD     AX,BX                  ; Add in the new digit.
SPACE1:   INC     SI                     ; Go on to the next character.
          LOOP    NUMLP1                 ; Loop back for more.
          MOV     BX,AX                  ; Store the answer in BX.
          JMP     NUMDN1
ERR1:     MOV     ERROR,X_FLD            ; Set up error.
NUMDN1:   RESTORE AX,CX,DX
          RET
CVRT      ENDP  
;----------------------------------------------------------------------------
CVRTDW    PROC    NEAR  
;  
; This procedure converts the 10-digit ASCII string pointed to by SI into  
; a double-word integer in DX:AX.  
;  
; The same rules apply here as in CVRT.  If you pass a non-numeric string,  
; CVRTDW will return 0 (this is not normally a legal record number).    
; Essentially, this procedure halts and returns what it currently has  
; calculated as soon as it finds either the end of the string or a non-  
; numeric character.  
;
          SAVE    BX,CX,DI
          SUB     AX,AX                  ; Clear out AX.
          SUB     DX,DX                  ; Clear out DX.
          MOV     CX,10                  ; Set loop counter for 10 digits.
NUMLP2:   CMP     BYTE PTR [SI],32       ; See if we got a space.
          JE      SPACE2                 ; If so, go on to next character.
          CMP     BYTE PTR [SI],'0'      ; See if digit less than '0'.
          JB      ERR2                   ; If so, quit.
          CMP     BYTE PTR [SI],'9'      ; See if digit > '9'
          JA      ERR2                   ; If so, quit.
          MOV     DI,TEN                 ; Get set for multiply.
          CALL    MULT                   ; Multiply DX:AX by DI.
          XOR     BH,BH                  ; Clear out BH.
          MOV     BL,[SI]                ; Get the next digit.
          SUB     BX,'0'                 ; Subtract ASCII offset.
          ADD     AX,BX                  ; Add in the new digit.
          ADC     DX,0                   ; Update DX if necessary.
SPACE2:   INC     SI                     ; Get to next character.
          LOOP    NUMLP2                 ; Go back for more.
          JMP     NUMDN2                 ; No error.
ERR2:     MOV     ERROR,X_REC            ; Send back error.
NUMDN2:   RESTORE BX,CX,DI
          RET
CVRTDW    ENDP  
;----------------------------------------------------------------------------
MULT      PROC    NEAR  
;  
; Multiply the number in DX:AX by the word in DI and put in DX:AX.  
;
          PUSH    DX                     ; Save current DX.
          MUL     DI                     ; Multiply AX by DI.
          MOV     BX,DX                  ; Move the carry portion into BX.
          POP     DX                     ; Restore the old DX.
          PUSH    AX                     ; Save the current AX.
          MOV     AX,DX                  ; Get the old DX ready to multiply.
          MUL     DI                     ; Multiply by DI.
          ADD     AX,BX                  ; Add on the carry from the first MUL.
          MOV     DX,AX                  ; Move the new high word into DX.
          POP     AX                     ; Get the current low word back.
          RET
MULT      ENDP  
;----------------------------------------------------------------------------
MEMTEST   PROC    NEAR  
;  
; This procedure looks through the 10-byte memo-field information string  
; looking for something besides spaces.  If it finds anything, it has  
; to assume that there is a memo field attached and sends this information  
; back.  If all spaces are found, there is no memo field.  
;
          SAVE    AX,CX,SI
          MOV     SI,OFFSET CHECK        ; Point SI to memo information.
          MOV     ISMEMO,'N'             ; Initialize to No.
          CALL    CVRTDW                 ; Convert memo pointer to double int.
          CMP     DX,0                   ; See if high word is 0.
          JNE     MDONE                  ; If not, we have a memo field.
          CMP     AX,0                   ; See if low word is 0.
          JE      ISDONE                 ; If it is, we don't have a memo field.
  MDONE:  MOV     ISMEMO,'Y'             ; Change ISMEMO to 'Y'.
  ISDONE: RESTORE AX,CX,SI
          RET
  MEMTEST ENDP  
;----------------------------------------------------------------------------
  CODE    ENDS
          END     START

